/*
 * algorithmONE FUTILEreleaseFOUR
 * Copyright (C) 1998 aempirei
 * Revised 29apr98
 * Published by The Ambient Empire
 * see futile.txt for license information
 */

#include <stdio.h>
#include <string.h>


/* Prototype Definitions */
void EVOLVE_KBOX(void);
void EVOLVE_SBOX(unsigned char);
void        SWAP(register unsigned char *, register unsigned char *);
void    SHOWHELP(void);
int   KEY_MASHER(char *);

static unsigned char  key[ 16] = { 0 }, sbox[256] = { 0 }, kbox[256] = { 0 };
static unsigned char pbox[256] = { /* The first 256 prime numbers MOD 256 */
   0x02, 0x03, 0x05, 0x07, 0x0B, 0x0D, 0x11, 0x13, 0x17, 0x1D,
    0x1F, 0x25, 0x29, 0x2B, 0x2F, 0x35, 0x3B, 0x3D, 0x43, 0x47,
    0x49, 0x4F, 0x53, 0x59, 0x61, 0x65, 0x67, 0x6B, 0x6D, 0x71,
    0x7F, 0x83, 0x89, 0x8B, 0x95, 0x97, 0x9D, 0xA3, 0xA7, 0xAD,
    0xB3, 0xB5, 0xBF, 0xC1, 0xC5, 0xC7, 0xD3, 0xDF, 0xE3, 0xE5,
    0xE9, 0xEF, 0xF1, 0xFB, 0x01, 0x07, 0x0D, 0x0F, 0x15, 0x19,
    0x1B, 0x25, 0x33, 0x37, 0x39, 0x3D, 0x4B, 0x51, 0x5B, 0x5D,
    0x61, 0x67, 0x6F, 0x75, 0x7B, 0x7F, 0x85, 0x8D, 0x91, 0x99,
    0xA3, 0xA5, 0xAF, 0xB1, 0xB7, 0xBB, 0xC1, 0xC9, 0xCD, 0xCF,
    0xD3, 0xDF, 0xE7, 0xEB, 0xF3, 0xF7, 0xFD, 0x09, 0x0B, 0x1D,
    0x23, 0x2D, 0x33, 0x39, 0x3B, 0x41, 0x4B, 0x51, 0x57, 0x59,
    0x5F, 0x65, 0x69, 0x6B, 0x77, 0x81, 0x83, 0x87, 0x8D, 0x93,
    0x95, 0xA1, 0xA5, 0xAB, 0xB3, 0xBD, 0xC5, 0xCF, 0xD7, 0xDD,
    0xE3, 0xE7, 0xEF, 0xF5, 0xF9, 0x01, 0x05, 0x13, 0x1D, 0x29,
    0x2B, 0x35, 0x37, 0x3B, 0x3D, 0x47, 0x55, 0x59, 0x5B, 0x5F,
    0x6D, 0x71, 0x73, 0x77, 0x8B, 0x8F, 0x97, 0xA1, 0xA9, 0xAD,
    0xB3, 0xB9, 0xC7, 0xCB, 0xD1, 0xD7, 0xDF, 0xE5, 0xF1, 0xF5,
    0xFB, 0xFD, 0x07, 0x09, 0x0F, 0x19, 0x1B, 0x25, 0x27, 0x2D,
    0x3F, 0x43, 0x45, 0x49, 0x4F, 0x55, 0x5D, 0x63, 0x69, 0x7F,
    0x81, 0x8B, 0x93, 0x9D, 0xA3, 0xA9, 0xB1, 0xBD, 0xC1, 0xC7,
    0xCD, 0xCF, 0xD5, 0xE1, 0xEB, 0xFD, 0xFF, 0x03, 0x09, 0x0B,
    0x11, 0x15, 0x17, 0x1B, 0x27, 0x29, 0x2F, 0x51, 0x57, 0x5D,
    0x65, 0x77, 0x81, 0x8F, 0x93, 0x95, 0x99, 0x9F, 0xA7, 0xAB,
    0xAD, 0xB3, 0xBF, 0xC9, 0xCB, 0xCF, 0xD1, 0xD5, 0xDB, 0xE7,
    0xF3, 0xFB, 0x07, 0x0D, 0x11, 0x17, 0x1F, 0x23, 0x2B, 0x2F,
    0x3D, 0x41, 0x47, 0x49, 0x4D, 0x53 };

int main(int argc, char *argv[]) {
   FILE *p_file, *c_file;
   char p_buf[1024], c_buf[1024];

   unsigned long i, t;
   register unsigned char p_byte = 0, c_byte = 0, l_byte = 0;
   register unsigned char j = 0, k = 0, x = 0, y = 0;

   /* Parameter Error control */

   if(argc != 5) {
	fprintf(stderr, "\nERROR: incorrect number of arguments");
	SHOWHELP(); return 0;
   }

   if(((argv[3][0]&'D') != 'D') && ((argv[3][0]&'E') != 'E')) {
      fprintf(stderr, "\nERROR: illegal switch %c",argv[3][0]);
      SHOWHELP(); return 0;
   }

   if (strlen(argv[4]) != 22) {
      fprintf(stderr, "\nERROR: incorrect password length");
      SHOWHELP(); return 0;
   }

   if(KEY_MASHER(argv[4]) == 1) {
      fprintf(stderr, "\nERROR: invalid character in password string");
      SHOWHELP(); return 0;
   }

   /* open INPUT and OUTPUT streams */
   if(argv[1][0] != '!') p_file = fopen(argv[1], "rb"); else p_file = stdin;
   if(argv[2][0] != '!') c_file = fopen(argv[2], "wb"); else c_file = stdout;

   /* setup I/O buffering */
   setvbuf(p_file, p_buf, _IOFBF, 1024);
   setvbuf(c_file, c_buf, _IOFBF, 1024);

   /* fopen error control */
   if(p_file == NULL) {
	fprintf(stderr, "ERROR: cannot open %s for input\n", argv[1]);
	return 0;
     }

   if(c_file == NULL) {
      fprintf(stderr, "ERROR: cannot open %s for output\n", argv[2]);
      return 0;
   }

   EVOLVE_KBOX(); /* (1) initialize KEY BOX */
   EVOLVE_KBOX(); /* (2) initialize KEY BOX */
   EVOLVE_KBOX(); /* (3) initialize KEY BOX */
   EVOLVE_KBOX(); /* (4) initialize KEY BOX */
   EVOLVE_SBOX(kbox[255]); /* (1) initialize SUB BOX */

   while(!feof(p_file)) { /* this is the main ENCRYPTION/DECRYPTION loop */

      t = (((unsigned long)pbox[0x00] *
	    (unsigned long)pbox[0x7F] *
	    (unsigned long)pbox[0xFF]) % 0x3000L) + 0x1000L;

      for(i = 0; i < t; i++) {
	 j++;
	 k  = kbox[j] + pbox[k];
	 x += sbox[((k & 15) << 4) | ((kbox[j] ^ pbox[j]) >> 4)];
	 y  = sbox[((j & 15) << 4) | ((kbox[k] ^ pbox[k]) >> 4)];

	 SWAP(&sbox[((j & 15) << 4) | (k & 15)],
	      &sbox[((x & 15) << 4) | (y & 15)]);

	 p_byte = fgetc(p_file); /* Read data byte */
	 if(feof(p_file)) break; /* Test for EOF   */
	 c_byte = p_byte ^ sbox[(j & 3) | (k & 12) | (x & 48) | (y & 192)];

	 fputc(c_byte, c_file); /* Write Data to file */
      }
      if(argv[3][0] == 'd') l_byte = c_byte;
      else l_byte = p_byte; /* Choose proper method (ENCRYPT/DECRYPT) */

      EVOLVE_KBOX();       /* Permuate KBOX */
      EVOLVE_SBOX(l_byte); /* Permuate SBOX */
   }

   fclose(p_file); /* flush buffer and close INPUT stream */
   fclose(c_file); /* flush buffer and close OUTPUT stream */

   return 0;
}

void EVOLVE_KBOX() {
   unsigned char i = 0, j = 0, _10h = 0;

   for(i = 0; i < 8; i++) {
      _10h = i * 32;

      for(j = 0; j < 16; j++) kbox[_10h + j +  0] = key[j];
      for(j = 0; j <  8; j++) kbox[_10h + j + 16] = key[ j << 1      ] + key[(j << 1) +  1];
      for(j = 0; j <  4; j++) kbox[_10h + j + 24] = key[(j << 1) + 16] + key[(j << 1) + 17];

      kbox[_10h + 28] = kbox[_10h + 24] + kbox[_10h + 25];
      kbox[_10h + 29] = kbox[_10h + 26] + kbox[_10h + 27];
      kbox[_10h + 30] = kbox[_10h + 28] + kbox[_10h + 29];
      kbox[_10h + 31] = 0;

      for(j = 0; j < 16; j++) kbox[_10h + 31] |= kbox[_10h + j] & (1 << j);

      for(j = 0; j < 31; j++) SWAP(&pbox[kbox[_10h + j]], &pbox[kbox[_10h + 31]]);

      for(j = 0; j < 16; j++) key[j] += pbox[kbox[_10h + 31]];
      for(j = 0; j < 16; j++) key[j] ^= key[(j + 1) % 16];
      for(j = 0; j < 16; j++) key[j] ^= key[(j + 1) % 16];
   }
}

/* This is a key-dependant S-BOX Permuation */
void EVOLVE_SBOX(unsigned char LAST) {
   unsigned char SWAPBOX = 0;
   unsigned int j = 0;

   for(j = 0; j < 256; j++) sbox[j] = (j ^ kbox[LAST]) + LAST;

   for(j = 0; j < 256; j++) {
      SWAPBOX =
	sbox[((pbox[SWAPBOX] & 15) << 4) | (j & 15)] +
	pbox[(SWAPBOX + j) & 0xFF] +
	kbox[j];

      SWAP(&sbox[pbox[j]], &sbox[kbox[j]]);
   }
}

/* Compress key to 128-bit */
int KEY_MASHER(char key_pointer[22]) {

   int j = 0;

   unsigned char i = 0;
   unsigned char cprs[256];
            char temp[ 22];

   strncpy(temp, key_pointer, 22);

   for(j = 0; j < 256; j++) cprs[j] = 0xFF;

   for(i = 'a'; i < 'z' + 1; i++) cprs[i] = i - 'a';
   for(i = 'A'; i < 'Z' + 1; i++) cprs[i] = i - 'A' + 26;
   for(i = '0'; i < '9' + 1; i++) cprs[i] = i - '0' + 52;
   cprs['+'] = 62;
   cprs['-'] = 63;

   for(j = 0; j < 22; j++) if(cprs[temp[j]] == 0xFF) return 1;

   for(i = 0; i < 5; i++) {
      key[i * 3    ] =  (cprs[temp[(i << 2)    ]]       << 2) |
			(cprs[temp[(i << 2) + 1]]       >> 4);
      key[i * 3 + 1] = ((cprs[temp[(i << 2) + 1]] & 15) << 4) |
			(cprs[temp[(i << 2) + 2]]       >> 2);
      key[i * 3 + 2] = ((cprs[temp[(i << 2) + 2]] &  3) << 6) |
			(cprs[temp[(i << 2) + 3]]           );
   }

   key[15] = (cprs[temp[20]] << 2) | (cprs[temp[21]] & 3);

   return 0;
}

/* Swaps CHAR type variables */
void SWAP(register unsigned char *a, register unsigned char *b) {
   register unsigned char c = *a;
   *a = *b;
   *b = c;
}

/* Display command-line help text. */
void SHOWHELP() {
   fprintf(stderr, "\n\n    ************************************************************************");
   fprintf(stderr, "\n    * algorithmONE  FUTILEreleaseFOUR               Ambient Empire 29apr98 *");
   fprintf(stderr, "\n    * Copyright (C) 1998 aempirei               see futile.txt for license *");
   fprintf(stderr, "\n    ************************************************************************\n\n");
   fprintf(stderr, "USAGE:  futile <source> <destination> <e/d> <password>\n\n");
   fprintf(stderr, "  <source>                    specifies file to encrypt or decrypt\n");
   fprintf(stderr, "  <destination>               specifies file to output data to\n");
   fprintf(stderr, "  <switch>                    d(ecrypt) or e(ncrypt)\n");
   fprintf(stderr, "  <password>                  22 character password\n");
   fprintf(stderr, "                              Valid Characters: A-Z a-z 0-9 `+' `-'\n\n");
   fprintf(stderr, "  ( for standard I/O devices and piping support, the '!'\n");
   fprintf(stderr, "    can be specified as the <source> and/or <destination> )\n");
}
